local MAJOR = "LibScriptableDisplayWidgetImage-1.0"
local MINOR = 14

assert(LibStub, MAJOR.." requires LibStub")
local WidgetImage = LibStub:NewLibrary(MAJOR, MINOR)
if not WidgetImage then return end
local LibWidget = LibStub("LibScriptableDisplayWidget-1.0", true)
assert(LibWidget, MAJOR .. " requires LibScriptableDisplayWidget-1.0")
local LibProperty = LibStub("LibScriptableDisplayProperty-1.0", true)
assert(LibProperty, MAJOR .. " requires LibScriptableDisplayProperty-1.0")
local LibTimer = LibStub("LibScriptableDisplayTimer-1.0", true)
assert(LibTimer, MAJOR .. " requires LibScriptableDisplayTimer-1.0")
local LibEvaluator = LibStub("LibScriptableDisplayEvaluator-1.0", true)
assert(LibEvaluator, MAJOR .. " requires LibScriptableDisplayEvaluator-1.0")
local LibBuffer = LibStub("LibScriptableDisplayBuffer-1.0", true)
assert(LibBuffer, MAJOR .. " requires LibScriptableDisplayBuffer-1.0")

local L = LibStub("LibScriptableDisplayLocale-1.0", true)
assert(L, MAJOR .. " requires LibScriptableDisplayLocale-1.0")
L = L.L

local pool = setmetatable({}, {__mode = "k"})

if not WidgetImage.__index then
	WidgetImage.__index = WidgetImage
end

WidgetImage.defaults = {
	speed = 500,
	repeating = true
}

local widgetType = {icon=true}

local function Split(str, delim)
	local start = 1
	local tbl = {}
	for i = 2, strlen(str) do
		if str:sub(i, i) == delim then
			tinsert(tbl, str:sub(start, i - 1))
			start = i + 1
		end
	end

	if start < strlen(str) then
		tinsert(tbl, str:sub(start, strlen(str)))
	end

	return tbl
end

-- Create a new LibScriptableDisplayWidgetImage object
-- @name WidgetImage.New
-- @usage WidgetImage:New(visitor, name, config, row, col, layer, errorLevel, timer)
-- @param visitor An LibScriptableDisplayCore-1.0 object, or provide your own
-- @param name A name for this widget
-- @param config This widget's settings
-- @param row This widget's row
-- @param col This widget's column
-- @param layer This widget's layer
-- @param errorLevel The errorLevel for this widget
-- @param callback The draw function
-- @param timer An optional timer. This should have a :Start() and :Stop().
-- @return A new LibScriptableDisplayWidgetImage widget
function WidgetImage:New(visitor, name, config, row, col, layer, errorLevel, callback, timer)
	assert(name, "WidgetImage requires a name.")
	assert(config, "Please provide the icon with a config")

	local obj = next(pool)

	if obj then
		pool[obj] = nil
		obj.__index = nil
	else
		obj = {}
		obj.options = {}
	end

	setmetatable(obj, self)

	obj.widget = LibWidget:New(obj, visitor, name, config, row, col, layer, widgetType, errorLevel)

	obj.config = config
	obj.Draw = callback
	obj.timer = timer

	obj.error = LibError:New(MAJOR .. ": " .. name, errorLevel)

	obj:Init()

	return obj
end

function WidgetImage:Init()

	self.width = self.config.width
	self.height = self.config.height

	self.speed = self.config.speed or self.defaults.speed

	self.fg_color = self.config.foreground or {self:FG_COL()}
	self.bg_color = self.config.background or {self:BG_COL()}

	self.index = 1
	self.repeating = self.config.repeating or self.defaults.repeating
	self.speed = self.config.speed or self.defaults.speed

	self.timer = self.timer or LibTimer:New("WidgetImage.timer " .. self.name, self.fps, self.repeating, self.Update, self, self.errorLevel)

	self.visible = LibProperty:New(self, self.visitor, self.name, self.config.visible, "return true", self.errorLevel);
	self.expr_max = LibProperty:New(self, self.visitor, self.name, self.config.expr_max, "return 0", self.errorLevel)
	self.expr_min = LibProperty:New(self, self.visitor, sefl.name, self.config.expr_min, "return 0", sefl.errorLevel)

	self.prescript = self.config.prescript
	self.script = self.config.script

	self.min = 0
	self.max = 0

	self.layers = config.layers or self.visitor.lcd.LAYERS or 1
	
	self.image = {}
	for i = 1, self.layers do
		tinsert(self.image, LibBuffer:New(self.name, self.width * self.height, self.bg_color, errorLevel))
	end

	--QObject::connect(visitor_->GetWrapper(), SIGNAL(_ResizeLCD(int, int, int, int)),
	--	this, SLOT(Resize(int, int, int, int)));

	return obj
end



-- Delete a LibScriptableDisplayWidgetImage object
-- @name LibScriptableDisplayWidgetImage.Del
-- @usage LibScriptableDisplayWidgetImage:Del([marq])
-- @param marq A LibScriptableDisplayWidgetImage object
function WidgetImage:Del(key)
	if not key then
		key = self
	end
	key.timer:Del()
	key.visible:Del()
	key.widget:Del()
	key.buffer:Del()
	key:Stop()
	pool[key] = true
end

-- Start a LibScriptableDisplayWidgetImage
-- @name LibScriptableDisplayWidgetImage.Start
-- @usage object:Start()
-- @return Nothing
function WidgetImage:Start()
	if(self.speed <= 0 or self.active) then
		return;
	end
	self:Clear()
	local val = LibEvaluator.ExecuteCode(self.environment, "WidgetImage.prescript", self.prescript)
	if not val and val ~= false or val then
		self.timer:Start();
		self.active = true
		self:Update();
	end
end

-- Stop a LibScriptableDisplayWidgetImage
-- @name LibScriptableDisplayWidgetImage.Stop
-- @usage object:Stop()
-- @return Nothing
function WidgetImage:Stop()
	self.timer:Stop()
	self.active = false
end

function WidgetImage:Resize(rows, cols, old_rows, old_cols)
	local yres = self.visitor.YRES;
	local xres = self.visitor.XRES;
	local r = self.row * yres / old_rows;
	local c = self.col * xres / old_cols;
	self.row = round(rows * r / yres);
	self.col = round(cols * c / xres);
	self:Update();
end

--[[
function WidgetImage:SetupChars()
	self.start = #self.visitor.lcd.specialChars
	for i, v in ipairs(self.icon) do
		tinsert(self.visitor.lcd.specialChars, v)
	end
	self.finish = #self.visitor.lcd.specialChars
end
]]

function WidgetImage:DrawVertLine(x, y1, y2, color)
	local y
	local p
	local image = self.image[self.index]

	color = color or self.fg_color or 0xff

	if(y1 < y2) then
		p = (y1+1)+x+1
		for y = y1, y2 do
			image.buffer[p] = color
			p = p + 1
		end
	elseif(y2 < y1) then
		p = (y2+1)+x+1
		for y = y2, y1 do
			image.buffer[p] = color
			p = p + 1
		end
	else
		image.buffer[(y1+1)+x+1] = color
	end
end

function WidgetImage:Clear()
	for n = 1, self.image:Size() do
		self.image[self.index].buffer[n] = self.bg_color or 0x00
	end
end

function WidgetImage:Fill(color)
	for n = 1, self.image[self.index]:Size() do
		self.image[self.index].buffer[n] = color or self.fg_color or 0xff
	end
end

local WidgetImage:FG_COL()
	return 0x00, 0x00, 0x00, 0xff
end

local WidgetImage:BG_COL()
	return 0xff, 0xff, 0xff, 0xff
end

local WidgetIMage:BL_COL()
	return 0xff, 0xff, 0xff, 0x00
end

local WidgetImage:NO_COL()
	return 0x00, 0x00, 0x00, 0x00
end

local WidgetImage:Blend(row, col)
	if row >= self.height or col >= self.width then return end
	local r, g, b, a = self:BL_COL()
	local o = self.layers
	for l = 1, self.layers do
		local p = self.image[l].buffer[row * lcd.LCOLS + col]
		if p.a == 255 then
			o = l
		end
	end
	for l = o, 1, -1 do
		local p = self.image[l].buffer[row * lcd.LCOLS + col]
		if p.a == 255 then
			r = p.r
			g = p.g
			b = p.b
		elseif p.a > 0 then
			r = (p.r * p.a + r * (255 - p.a)) / 255
			g = (p.g * p.a + g * (255 - p.a)) / 255
			b = (p.b * p.a + b * (255 - p.a)) / 255
		end
	end
	return r, g, b, 0xff
end

local function incrementIndex(self)
	self.index = self.index + 1
	if(self.index > #self.image) then
		self.index = 1;
	end
end

function WidgetImage:Update()
	if not self.active then return end
	self.visible:Eval()
	self.expr_max:Eval()
	self.expr_min:Eval()

	local val = LibEvaluator.ExecuteCode(self.environment, "WidgetImage.script", self.script)

	if type(val) == "number" then

		local max, min
		if( self.expr_min.is_valid ) then
			self.expr_min:Eval();
			min = self.expr_min:P2N();
			assert(type(min) == "number", format("%s: expr_min is not a number value (%s)", self.name, type(min)))
		else
			min = self.min;
			if( val < min ) then
				min = val;
			end
		end

		if( self.expr_max.is_valid ) then
			self.expr_max:Eval();
			max = self.expr_max:P2N();
			assert(type(max) == "number", format("%s: expr_max is not a number value (%s)", self.name, type(max)))
		else
			max = self.max;
			if( val > max ) then
				max = val;
			end
		end

		self.min = min;
		self.max = max;

		if( max > min ) then
			self.val = (val - min ) / (max - min);
		else
			self.val = 0.0;
		end

		self.index = floor(#self.image * self.val + 0.5) + 1
		if self.index < 1 then
			self.index = 1
		elseif self.index > #self.icon then
			self.index = #self.icon
		end
	elseif type(val) == "boolean" then
		if val then
			incrementIndex(self)
		end
	end

	self:Draw(val);
end

function WidgetImage:GetOptions(visitor, db)
	local defaults = WidgetImage.defaults
	local options = {
		enabled = {
			name = "Enabled",
			desc = "Whether this icon is enabled or not",
			type = "toggle",
			get = function() return db.enabled end,
			set = function(info, v) db.enabled = v; db["enabledDirty"] = true end,
			order = 1
		}
	}
	options.widget = LibWidget:GetOptions(visitor, db)
	return options
end

